#include <ctype.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>

#include "list.h"

struct ctrie_node
   {
   char * key;
   struct ctrie_node * parent;
   struct ctrie_node * child;
   struct ctrie_node * next_sibling;
   struct ctrie_node * prev_sibling;
   void * data;
   };

struct ctrie
   {
   struct ctrie_node * root;
   size_t size;
   size_t count;
   };


static struct list * subkey_list = NULL;
static size_t subkey_list_limit = 0;

struct ctrie_node * ctrie_search_children(struct ctrie_node * node, char k);
void ctrie_visit_node(struct ctrie * ctrie, struct ctrie_node * node, void (*op)(char * k, void * d), char * key);
void ctrie_print_node(struct ctrie_node * node);
void ctrie_print_node(struct ctrie_node * node);
void ctrie_node_dump_preorder(struct ctrie * ctrie, struct ctrie_node * node, int depth);
void ctrie_node_dump_raw(struct ctrie_node * node, int depth);
void ctrie_subtree_list_push(char * k, void * d);
struct ctrie_node * ctrie_find_node(struct ctrie * ctrie, char * key);

struct ctrie_node * ctrie_search_children_longest_match(struct ctrie_node * node, char * k);

#undef MAX
#define MAX (a, b) ((a) > (b) ? (a) : (b))

////////////////////////////////////////////////////////////////////////////////
struct ctrie *
ctrie_new()
   {
   struct ctrie * ctrie = malloc(sizeof(struct ctrie));
   if (NULL == ctrie)
      return NULL;

   ctrie->root = malloc(sizeof(struct ctrie_node));
   if (NULL == ctrie->root)
      return NULL;

   ctrie->root->parent = ctrie->root->child = ctrie->root->next_sibling = NULL;
   ctrie->root->key = NULL;
   ctrie->root->data = NULL;
   ctrie->size = ctrie->count = 0;

   return ctrie;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_delete(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return;

   free(ctrie);
   }

////////////////////////////////////////////////////////////////////////////////
size_t
ctrie_size(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return 0;
   return ctrie->size;
   }

////////////////////////////////////////////////////////////////////////////////
size_t
ctrie_count(struct ctrie * ctrie)
   {
   if (NULL == ctrie)
      return 0;
   return ctrie->count;
   }

////////////////////////////////////////////////////////////////////////////////
char *
ctrie_insert(struct ctrie * ctrie, char * key, void * data)
   {
   if ( (NULL == ctrie) || (NULL == key) )
      return NULL;

   struct ctrie_node * curr = ctrie->root;
   struct ctrie_node * result;

   size_t len = strlen(key);
   for (int i = 0; i <= len; ++i)
      {
      result = ctrie_search_children_longest_match(curr, key+i);
      if (NULL == result)
	 {
	 // current char is not in the ctrie yet, allocate a new node and attach it to the children
	 struct ctrie_node * new = malloc(sizeof(struct ctrie_node));
	 if (NULL == new)
	    return NULL;
	 new->key = tolower(key[i]);
	 new->child = new->next_sibling = new->prev_sibling = NULL;
	 new->data = NULL;

	 if (NULL == curr->child)
	    curr->child = new;
	 else
	    {
	    struct ctrie_node * p = curr->child;
	    while (p->next_sibling)
	       p = p->next_sibling;
	    new->prev_sibling = p;
	    p->next_sibling = new;
	    }

	 new->parent = curr;
	 curr = new;
	 ctrie->size++;
	 }
      else
	 curr = result;
      }

   // At this point, curr is either pointing to a new \0 node with a NULL data pointer,
   // or an existing \0 node with a valid data pointer (in which case we were asked to "insert" 
   // a key that was already present in the ctrie.
   // For now I am simply ignoring collisions.  We will update the data.

   curr->data = data;

   ctrie->count++;

   return key;
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_search_children_longest_match(struct ctrie_node * node, char * k)
   {
   struct ctrie_node * p = node->child;
   if (NULL == p)
      return NULL;

   size_t klen = strlen(k);
   struct ctrie_node * longest = NULL;
   size_t length = 0;
   while (NULL != p)
      {
      size_t len = strlen(p->key);
      while (len > 0)
	 {
	 if ( (0 == strncmp(p->key, k, MAX(len, klen))) &&
	      (length < MAX(len, klen)) )
	    {
	    longest = p;
	    length = MAX(len, klen);
	    break;
	    }
	 len--;
	 }
      p = p->next_sibling;
      }

   return p;
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_search_children(struct ctrie_node * node, char k)
   {
   struct ctrie_node * p = node->child;
   if (NULL == p)
      return NULL;

   while ( (NULL != p) && (p->key != tolower(k)) )
      p = p->next_sibling;
   return p;
   }

////////////////////////////////////////////////////////////////////////////////
struct ctrie_node *
ctrie_find_node(struct ctrie * ctrie, char * key)
   {
   if ( (NULL == ctrie) || (NULL == key) )
      return NULL;

   struct ctrie_node * node = ctrie->root;
   size_t len = strlen(key);
   for (int i = 0; i <= len; i++)
      {
      node = ctrie_search_children(node, key[i]);
      if (NULL == node)
	 return NULL;
      if ('\0' == key[i])
	 return node;
      }

   // Should never get here
   return NULL;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ctrie_find(struct ctrie * ctrie, char * key)
   {
   struct ctrie_node * node = ctrie_find_node(ctrie, key);
   if (NULL == node)
      return NULL;
   return node->data;
   }

////////////////////////////////////////////////////////////////////////////////
void *
ctrie_remove(struct ctrie * ctrie, char * key)
   {
   if ( (NULL == ctrie) || (NULL == key) )
      return NULL;

   struct ctrie_node * node = ctrie_find_node(ctrie, key);
   if (NULL == node)
      return NULL;

   void * data = node->data;

   while ( (NULL == node->prev_sibling) && (NULL == node->prev_sibling) ) 
      {
      struct ctrie_node * p = node;
      node = node->parent;
      free(p);
      ctrie->size--;
      }

   if (node == ctrie->root)
      return data;

   if (node->prev_sibling)
      node->prev_sibling->next_sibling = node->next_sibling;
   else
      node->parent->child = node->next_sibling;

   if (node->next_sibling)
      node->next_sibling->prev_sibling = node->prev_sibling;

   free(node);

   ctrie->count--;

   return data;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_walk_keys(struct ctrie * ctrie, void (*op)(char * k, void * d))
   {
   if (NULL == ctrie)
      return;
   ctrie_visit_node(ctrie, ctrie->root, op, "");
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_visit_node(struct ctrie * ctrie, struct ctrie_node * node, void (*op)(char * k, void * d), char * key)
   {
   if ( (node->key == '\0') && (node != ctrie->root) )
      op(key, node->data);
   else
      {
      size_t len = strlen(key);
      char * k = malloc(len+2);
      if (NULL == k)
	 return;
      strcpy(k, key);
      k[len] = node->key;
      k[len+1] = '\0';
      struct ctrie_node * p = node->child;
      while (NULL != p)
	 {
	 ctrie_visit_node(ctrie, p, op, k);
	 p = p->next_sibling;
	 }
      free(k);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void ctrie_dump_raw(struct ctrie * ctrie)
   { 
   if (NULL == ctrie)
      return;
  
   int depth = -1;
   ctrie_node_dump_raw(ctrie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_node_dump_raw(struct ctrie_node * node, int depth)
   {
   for (int i=0; i < depth; i++)
      printf(" ");
   ctrie_print_node(node);
   if (node->child)
      ctrie_node_dump_raw(node->child, depth+1); 
   if (NULL != node->next_sibling)
      ctrie_node_dump_raw(node->next_sibling, depth+1);
   }

////////////////////////////////////////////////////////////////////////////////
void ctrie_dump(struct ctrie * ctrie)
   { 
   if (NULL == ctrie)
      return;
  
   int depth = -1;
   ctrie_node_dump_preorder(ctrie, ctrie->root, depth);
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_node_dump_preorder(struct ctrie * ctrie, struct ctrie_node * node, int depth)
   {
   printf("%c", node->key);
   if ( (node != ctrie->root) && ('\0' == node->key) )
      printf("\n");
   if (node->child)
      ctrie_node_dump_preorder(ctrie, node->child, depth+1);
   if (NULL != node->next_sibling)
      {
      for (int i=0; i < depth; i++)
	 printf(" ");
      ctrie_node_dump_preorder(ctrie, node->next_sibling, depth);
      }
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_print_node(struct ctrie_node * node)
   {
   printf("%p\tchild %p\tnext_sibling %p\tkey %c\tdata %p\n", (void *) node, (void *) node->child, 
	  (void *) node->next_sibling, node->key, node->data);
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_subtree_list_push(char * k, void * d)
   {
   if (NULL == subkey_list)
      return;

   if ( (0 == subkey_list_limit) || (list_size(subkey_list) < subkey_list_limit) )
      list_push_front(subkey_list, k);
   }

////////////////////////////////////////////////////////////////////////////////
// Not reentrant since it uses the package static subkeys_list variable.
// However, that variable is only used while building a subkey list, so you can
// get more than one subkey if you want.  Just don't try to get them simultaneously.
struct list *
ctrie_get_subkeys(struct ctrie * ctrie, char * prefix, size_t limit)
   {
   if ( (NULL == ctrie) || (NULL == prefix) )
      return NULL;

   struct ctrie_node * node = ctrie_find_node(ctrie, prefix);
   if (NULL == node)
      return NULL;

   subkey_list = list_new();
   if (NULL == subkey_list)
      return NULL;
   subkey_list_limit = limit;

   size_t len = strlen(prefix);
   char * pref = malloc(len+1);
   if (NULL == pref)
      return NULL;
   strncpy(pref, prefix, len-1);
   pref[len-1] = '\0';

   ctrie_visit_node(ctrie, node->parent, ctrie_subtree_list_push, pref);
   free(pref);

   return subkey_list;
   }

////////////////////////////////////////////////////////////////////////////////
void
ctrie_free_subkey_list(struct list * list)
   {
   if (NULL == list)
      return;

   struct list_iterator * it = list_begin(list);
   void * data;
   while (data = list_remove(it))
      free(data);

   list_delete(list);
   }
